from typing import ClassVar

from ixmp4.core.exceptions import IxmpError
from ixmp4.data import types
from ixmp4.data.abstract.annotations import HasIdFilter, HasNameFilter, HasRunIdFilter
from ixmp4.db.filters import BaseFilter

from .. import mixins
from ..base import BaseModel as RootBaseModel
from ..base import (
    BulkDeleter,
    BulkUpserter,
    Creator,
    Deleter,
    Enumerator,
    Lister,
    Retriever,
    Selecter,
    Tabulator,
)


class BaseModel(RootBaseModel, mixins.HasCreationInfo):
    # NOTE: only subclasses storing data actually define this!
    DataInvalid: ClassVar[type[IxmpError]]

    __abstract__ = True

    name: types.Name


class EnumerateKwargs(HasIdFilter, HasNameFilter, HasRunIdFilter, total=False):
    _filter: BaseFilter



from typing import Any

from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.data.db.run.model import Run
from ixmp4.db import Session, filters, typing_column, utils

from . import Model


class BaseIamcFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    def join_datapoints(
        self, exc: db.sql.Select[tuple[Model]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Model]]:
        if not utils.is_joined(exc, Run):
            # This looks like any Select coming in here must have Model in it
            exc = exc.join(Run, onclause=Run.model__id == Model.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)

        return exc


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(
        self, exc: db.sql.Select[tuple[Model]], session: Session | None = None
    ) -> db.sql.Select[tuple[Model]]:
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, Run.model)
        return exc

    model: base.ModelFilter | None = filters.Field(default=None, exclude=True)


class IamcModelFilter(
    base.ModelFilter,
    BaseIamcFilter,
    metaclass=filters.FilterMeta,
):
    region: base.RegionFilter | None = filters.Field(None)
    variable: base.VariableFilter | None = filters.Field(None)
    unit: base.UnitFilter | None = filters.Field(None)
    run: RunFilter = filters.Field(
        default=RunFilter(id=None, version=None, is_default=True)
    )

    _remote_filters = {"run", "region", "variable", "unit"}
    _remote_path = [
        {
            "target_model": Run,
            "fk_attr": "model__id",
            "source_model": Model,
            "pk_attr": "id",
        },
        {
            "target_model": TimeSeries,
            "fk_attr": "run__id",
            "source_model": Run,
            "pk_attr": "id",
        },
    ]

    def join(
        self, exc: db.sql.Select[tuple[Model]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Model]]:
        if self._should_use_subquery_optimization():
            return exc

        return super().join_datapoints(exc, session)


class ModelFilter(base.ModelFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    iamc: IamcModelFilter | filters.Boolean

    def filter_iamc(
        self,
        exc: db.sql.Select[tuple[Model]],
        c: typing_column[Any],
        v: bool | None,
        session: db.Session | None = None,
    ) -> db.sql.Select[tuple[Model]]:
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Model.id), session)
            exc = exc.where(~Model.id.in_(ids))
            return exc

    def join(
        self, exc: db.sql.Select[tuple[Model]], session: Session | None = None
    ) -> db.sql.Select[tuple[Model]]:
        return exc



from collections.abc import Iterable
from typing import TYPE_CHECKING

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.abstract.annotations import HasUnitIdFilter
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import ScalarDocsRepository
from .model import Scalar


class EnumerateKwargs(base.EnumerateKwargs, HasUnitIdFilter, total=False): ...


class ScalarRepository(
    base.Creator[Scalar],
    base.Deleter[Scalar],
    base.Retriever[Scalar],
    base.Enumerator[Scalar],
    abstract.ScalarRepository,
):
    model_class = Scalar

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ScalarDocsRepository(*args)

        from .filter import OptimizationScalarFilter

        self.filter_class = OptimizationScalarFilter

    def add(self, name: str, value: float | int, unit_name: str, run_id: int) -> Scalar:
        unit_id = self.backend.units.get(unit_name).id
        scalar = Scalar(
            name=name,
            value=value,
            unit__id=unit_id,
            run__id=run_id,
        )
        self.session.add(scalar)
        return scalar

    @guard("edit")
    def delete(self, id: int) -> None:
        return super().delete(id=id)

    @guard("view")
    def get(self, run_id: int, name: str) -> Scalar:
        exc = db.select(Scalar).where(
            (Scalar.name == name) & (Scalar.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Scalar.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Scalar:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Scalar.NotFound(id=id)

        return obj

    @guard("edit")
    def create(self, name: str, value: float, unit_name: str, run_id: int) -> Scalar:
        return super().create(
            name=name, value=value, unit_name=unit_name, run_id=run_id
        )

    @guard("edit")
    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        exc = db.update(Scalar).where(
            Scalar.id == id,
        )

        if value is not None:
            exc = exc.values(value=value)
        if unit_id is not None:
            exc = exc.values(unit__id=unit_id)

        self.session.execute(exc)
        self.session.commit()
        return self.get_by_id(id)

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> Iterable[Scalar]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from collections.abc import Generator
from typing import TYPE_CHECKING, Any, List, Literal, Sequence, cast

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.db.optimization.equation.model import (
    EquationIndexsetAssociation,
)
from ixmp4.data.db.optimization.equation.repository import EquationRepository
from ixmp4.data.db.optimization.parameter.model import (
    ParameterIndexsetAssociation,
)
from ixmp4.data.db.optimization.parameter.repository import ParameterRepository
from ixmp4.data.db.optimization.table.model import TableIndexsetAssociation
from ixmp4.data.db.optimization.table.repository import TableRepository
from ixmp4.data.db.optimization.variable.model import (
    VariableIndexsetAssociation,
)
from ixmp4.data.db.optimization.variable.repository import VariableRepository

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

import logging

import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard

from .. import base, utils
from .docs import IndexSetDocsRepository
from .model import IndexSet, IndexSetData

log = logging.getLogger(__name__)


class IndexSetRepository(
    base.Creator[IndexSet],
    base.Deleter[IndexSet],
    base.Retriever[IndexSet],
    base.Enumerator[IndexSet],
    abstract.IndexSetRepository,
):
    model_class = IndexSet

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = IndexSetDocsRepository(*args)

        from .filter import OptimizationIndexSetFilter

        self.filter_class = OptimizationIndexSetFilter

        self._linked_columns_lookup = {
            "table": (
                TableIndexsetAssociation.table__id,
                TableIndexsetAssociation.indexset__id,
            ),
            "parameter": (
                ParameterIndexsetAssociation.parameter__id,
                ParameterIndexsetAssociation.indexset__id,
            ),
            "equation": (
                EquationIndexsetAssociation.equation__id,
                EquationIndexsetAssociation.indexset__id,
            ),
            "variable": (
                VariableIndexsetAssociation.variable__id,
                VariableIndexsetAssociation.indexset__id,
            ),
        }

    def add(self, run_id: int, name: str) -> IndexSet:
        indexset = IndexSet(run__id=run_id, name=name)
        self.session.add(indexset)
        return indexset

    @guard("view")
    def get(self, run_id: int, name: str) -> IndexSet:
        exc = db.select(IndexSet).where(
            (IndexSet.name == name) & (IndexSet.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise IndexSet.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> IndexSet:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise IndexSet.NotFound(id=id)

        return obj

    @guard("edit")
    def create(self, run_id: int, name: str) -> IndexSet:
        return super().create(run_id=run_id, name=name)

    @guard("edit")
    def delete(self, id: int) -> None:
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> list[IndexSet]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs).rename(columns={"_data_type": "data_type"})

    @guard("edit")
    def add_data(
        self, id: int, data: float | int | str | List[float] | List[int] | List[str]
    ) -> None:
        _data = data if isinstance(data, list) else [data]

        if len(_data) == 0:
            return  # nothing to be done

        indexset = self.get_by_id(id=id)

        bulk_insert_enabled_data = [{"value": str(d)} for d in _data]
        try:
            self.session.execute(
                db.insert(IndexSetData).values(indexset__id=id),
                bulk_insert_enabled_data,
            )
        except db.IntegrityError as e:
            self.session.rollback()
            raise indexset.DataInvalid from e

        # Due to _data's limitation above, __name__ will always be that
        indexset._data_type = cast(
            Literal["float", "int", "str"], type(_data[0]).__name__
        )

        self.session.commit()

    @guard("edit")
    def remove_data(
        self,
        id: int,
        data: float | int | str | List[float] | List[int] | List[str],
        remove_dependent_data: bool = True,
    ) -> None:
        if not bool(data):
            return

        _data = [str(d) for d in data] if isinstance(data, list) else [str(data)]

        if remove_dependent_data:
            repos: dict[
                Literal["table", "parameter", "equation", "variable"],
                TableRepository
                | ParameterRepository
                | EquationRepository
                | VariableRepository,
            ] = {
                "table": self.backend.optimization.tables,
                "parameter": self.backend.optimization.parameters,
                "equation": self.backend.optimization.equations,
                "variable": self.backend.optimization.variables,
            }
            for kind, ids in self.find_all_linked_item_ids(id=id):
                self.remove_invalid_linked_data(
                    repo=repos[kind],
                    ids=ids,
                    indexset_name=self.get_by_id(id=id).name,
                    data=_data,
                )

        result = self.session.execute(
            db.delete(IndexSetData).where(
                IndexSetData.indexset__id == id,
                IndexSetData.value.in_(_data),
            )
        )

        if result.rowcount == 0:
            log.info(f"No data were removed! Are {data} registered to IndexSet {id}?")
            return None
        elif result.rowcount != len(_data):
            log.info(f"Not all items in `data` were registered for IndexSet {id}!")

        # Expire session to refresh IndexSets stored in it
        self.session.commit()

    def _find_linked_item_ids(
        self, id: int, item_kind: Literal["table", "parameter", "equation", "variable"]
    ) -> Sequence[int]:
        """Finds all items of `item_kind` linked to an IndexSet in `session.

        Parameters
        ----------
        id : int
            The id of the IndexSet we are looking for.
        item_kind : Literal["table", "parameter", "equation", "variable"]
            The type of item we are looking for.

        Returns
        -------
        list of int
            A list of ids of items of `item_kind` linked to the IndexSet.
        """

        column_clause, compare_column = self._linked_columns_lookup[item_kind]

        statement = db.select(column_clause).where(compare_column == id)

        return self.session.scalars(statement).all()

    @guard("view")
    def find_all_linked_item_ids(
        self, id: int
    ) -> Generator[
        tuple[Literal["table", "parameter", "equation", "variable"], Sequence[int]],
        Any,
        None,
    ]:
        """Finds all optimization items linked to an IndexSet.

        This is done by iterating over all possible kinds and yielding the ids of linked
        items.

        Parameters
        ----------
        id : int
            The id of the IndexSet we are looking for.

        Yields
        ------
        (item kind, list of ids)
            A tuple with the item kind being one of
            {'table', 'parameter', 'equation', 'variable'} and a list of integer ids
            representing linked items.
        """
        item_kinds: set[Literal["table", "parameter", "equation", "variable"]] = {
            "table",
            "parameter",
            "equation",
            "variable",
        }
        for kind in item_kinds:
            yield (kind, self._find_linked_item_ids(id=id, item_kind=kind))

    @guard("edit")
    def remove_invalid_linked_data(
        self,
        repo: TableRepository
        | ParameterRepository
        | EquationRepository
        | VariableRepository,
        ids: Sequence[int],
        indexset_name: str,
        data: List[str],
    ) -> None:
        """Remove invalid data from linked optimization items.

        Parameters
        ----------
        repo : TableRepository | ParameterRepository | EquationRepository | VariableRepository
            The repository including the linked items.
        ids : Sequence[int]
            The IDs of items linked to the IndexSet with `indexset_name`.
        indexset_name : str
            The name of the IndexSet from which data is to be removed.
        data : list[str]
            The data to be removed from `indexset_name` in str format.
        """  # noqa: E501
        for item in repo.list(id__in=ids):
            # Convert existing data for manipulation
            df = pd.DataFrame(item.data)

            if df.empty:
                continue  # nothing to do

            # Identify column/dimension names to target; item and name must be linked
            columns = utils._find_columns_linked_to_indexset(
                item=item, name=indexset_name
            )

            # Prepare data template to exclude
            invalid_data = {columns[i]: data for i in range(len(columns))}

            # Prepare stored data to remove
            # NOTE if any linked column is not of type str, this may be incorrect
            remove_data = df[df[columns].isin(invalid_data).any(axis=1)]

            repo.remove_data(item.id, remove_data)
